import unittest
import io

from ..scanner import preview_parse, filter_properties

test_ows = b"""\
<?xml version="1.0" encoding="utf-8"?>
<scheme title="Football" description="On this sunday" version="2.0" >
 <nodes>
  <node id="0" name="A" position="(152.0, 158.0)" />
  <node id="1" name="B" position="(316.0, 163.0)" />
  <node id="2" name="C" position="(166.0, 265.0)" />
  <node id="3" name="D" position="(459.0, 158.0)" />
 </nodes>
 <links>
  <link enabled="true" id="0" sink_channel="B.1" sink_node_id="1" source_channel="A.1" source_node_id="0" />
  <link enabled="true" id="1" sink_channel="B.2" sink_node_id="1" source_channel="B.1" source_node_id="2" />
  <link enabled="true" id="2" sink_channel="C.1" sink_node_id="3" source_channel="B.1" source_node_id="1" />
 </links>
 <node_properties>
  <properties format="pickle" node_id="0">random garbage</properties>
  <properties format="pickle" node_id="1">random garbage</properties>
 </node_properties>
</scheme>
"""


class TestPreviewParse(unittest.TestCase):
    def test_filter_properties(self):
        stream = io.BytesIO(test_ows)
        filtered = filter_properties(stream)
        self.assertNotIn(b'<properties>', filtered)
        self.assertEqual(filtered.count(b'<node '), 4)

    def test_preview_parse(self):
        stream = io.BytesIO(test_ows)
        a, b, c = preview_parse(stream)
        assert a == "Football"
        assert b == "On this sunday"
        assert c == ""



from AnyQt.QtWidgets import QAction, QToolButton

from .. import test
from ..toolgrid import ToolGrid


class TestToolGrid(test.QAppTestCase):
    def test_tool_grid(self):
        w = ToolGrid()

        w.show()
        self.app.processEvents()

        def buttonsOrderedVisual():
            # Process layout events so the buttons have right positions
            self.app.processEvents()
            buttons = w.findChildren(QToolButton)
            return list(sorted(buttons, key=lambda b: (b.y(), b.x())))

        def buttonsOrderedLogical():
            return list(map(w.buttonForAction, w.actions()))

        def assertOrdered():
            self.assertSequenceEqual(buttonsOrderedLogical(),
                                     buttonsOrderedVisual())

        action_a = QAction("A", w)
        action_b = QAction("B", w)
        action_c = QAction("C", w)
        action_d = QAction("D", w)

        w.addAction(action_b)
        w.insertAction(0, action_a)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b])
        assertOrdered()

        w.addAction(action_d)
        w.insertAction(action_d, action_c)

        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_c, action_d])
        assertOrdered()

        w.removeAction(action_c)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_d])

        assertOrdered()

        w.removeAction(action_a)
        self.assertSequenceEqual(w.actions(),
                                 [action_b, action_d])

        assertOrdered()

        w.insertAction(0, action_a)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_d])

        assertOrdered()

        w.setColumnCount(2)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_d])

        assertOrdered()

        w.insertAction(2, action_c)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_c, action_d])
        assertOrdered()

        w.clear()
        # test no 'before' action edge case
        w.insertAction(0, action_a)
        self.assertIs(action_a, w.actions()[0])
        w.insertAction(1, action_b)
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b])

        w.clear()
        w.setActions([action_a, action_b, action_c, action_d])
        self.assertSequenceEqual(w.actions(),
                                 [action_a, action_b, action_c, action_d])
        assertOrdered()

        triggered_actions = []

        def p(action):
            print(action.text())

        w.actionTriggered.connect(p)
        w.actionTriggered.connect(triggered_actions.append)
        action_a.trigger()

        w.show()
        self.qWait()



"""
Basic Qt testing framework
==========================
"""
import unittest
import gc
from typing import Callable, Any

from AnyQt.QtWidgets import QApplication, QWidget
from AnyQt.QtCore import (
    QCoreApplication, QTimer, QStandardPaths, QPoint, Qt, QMimeData, QPointF
)
from AnyQt.QtGui import (
    QMouseEvent, QDragEnterEvent, QDropEvent, QDragMoveEvent, QDragLeaveEvent,
    QContextMenuEvent
)
from AnyQt.QtTest import QTest
from AnyQt.QtCore import PYQT_VERSION

DEFAULT_TIMEOUT = 50


class QCoreAppTestCase(unittest.TestCase):
    _AppClass = QCoreApplication
    app = None  # type: QCoreApplication

    __appdomain = ""
    __appname = ""

    @classmethod
    def setUpClass(cls):
        super(QCoreAppTestCase, cls).setUpClass()
        QStandardPaths.setTestModeEnabled(True)
        app = cls._AppClass.instance()
        if app is None:
            app = cls._AppClass([])
        cls.app = app
        cls.__appname = cls.app.applicationName()
        cls.__appdomain = cls.app.organizationDomain()
        cls.app.setApplicationName("orangecanvas.testing")
        cls.app.setOrganizationDomain("biolab.si")

    def setUp(self):
        super(QCoreAppTestCase, self).setUp()

    def tearDown(self):
        super(QCoreAppTestCase, self).tearDown()

    @classmethod
    def tearDownClass(cls):
        gc.collect()
        cls.app.setApplicationName(cls.__appname)
        cls.app.setOrganizationDomain(cls.__appdomain)
        cls.app.sendPostedEvents(None, 0)
        # Keep app instance alive between tests with PyQt5 5.14.0 and later
        if PYQT_VERSION <= 0x050e00:
            cls.app = None
        super(QCoreAppTestCase, cls).tearDownClass()
        QStandardPaths.setTestModeEnabled(False)

    @classmethod
    def qWait(cls, timeout=DEFAULT_TIMEOUT):
        QTest.qWait(timeout)

    @classmethod
    def singleShot(cls, timeout: int, slot: 'Callable[[], Any]'):
        QTimer.singleShot(timeout, slot)


class QAppTestCase(QCoreAppTestCase):
    _AppClass = QApplication
    app = None  # type: QApplication


def mouseMove(widget, buttons, modifier=Qt.NoModifier, pos=QPoint(), delay=-1):
    # type: (QWidget, Qt.MouseButtons, Qt.KeyboardModifier, QPoint, int) -> None
    """
    Like QTest.mouseMove, but with `buttons` and `modifier` parameters.

    Parameters
    ----------
    widget : QWidget
    buttons: Qt.MouseButtons
    modifier : Qt.KeyboardModifiers
    pos : QPoint
    delay : int
    """
    if pos.isNull():
        pos = widget.rect().center()
    me = QMouseEvent(
        QMouseEvent.MouseMove, QPointF(pos), QPointF(widget.mapToGlobal(pos)),
        Qt.NoButton, buttons, modifier
    )
    if delay > 0:
        QTest.qWait(delay)

    QCoreApplication.sendEvent(widget, me)


def contextMenu(widget: QWidget, pos: QPoint, delay=-1) -> None:
    """
    Simulates a contextMenuEvent on the widget.
    """
    ev = QContextMenuEvent(
        QContextMenuEvent.Mouse, pos, widget.mapToGlobal(pos)
    )
    if delay > 0:
        QTest.qWait(delay)
    QCoreApplication.sendEvent(widget, ev)


def dragDrop(
        widget: QWidget, mime: QMimeData, pos: QPoint = QPoint(),
        action=Qt.CopyAction, buttons=Qt.LeftButton, modifiers=Qt.NoModifier
) -> bool:
    """
    Simulate a drag/drop interaction on the `widget`.

    A `QDragEnterEvent`, `QDragMoveEvent` and `QDropEvent` are created and
    dispatched to the `widget`. However if any of the `QDragEnterEvent` or
    `QDragMoveEvent` are not accepted, a `QDragLeaveEvent` is dispatched
    to 'reset' the widget state before this function returns `False`

    Parameters
    ----------
    widget: QWidget
        The target widget.
    mime: QMimeData
        The mime data associated with the drag/drop.
    pos: QPoint
        Position of the drop
    action: Qt.DropActions
        Type of acceptable drop actions
    buttons: Qt.MouseButtons:
        Pressed mouse buttons.
    modifiers: Qt.KeyboardModifiers
        Pressed keyboard modifiers.

    Returns
    -------
    state: bool
        Were the events accepted.

    See Also
    --------
    QDragEnterEvent, QDropEvent
    """
    if pos.isNull():
        pos = widget.rect().center()

    ev = QDragEnterEvent(pos, action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)

    ev = QDragMoveEvent(pos, action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)

    if not ev.isAccepted():
        QApplication.sendEvent(widget, QDragLeaveEvent())
        return False

    ev = QDropEvent(QPointF(pos), action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)
    return ev.isAccepted()


def dragEnterLeave(
        widget: QWidget, mime: QMimeData, pos=QPoint(),
        action=Qt.CopyAction, buttons=Qt.LeftButton, modifiers=Qt.NoModifier
) -> None:
    """
    Simulate a drag/move/leave interaction on the `widget`.

    A QDragEnterEvent, QDragMoveEvent and a QDragLeaveEvent are created
    and dispatched to the widget.
    """
    if pos.isNull():
        pos = widget.rect().center()

    ev = QDragEnterEvent(pos, action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)

    ev = QDragMoveEvent(
        pos, action, mime, buttons, modifiers, QDragMoveEvent.DragMove
    )
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)

    ev = QDragLeaveEvent()
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)
    return



"""
Test for DynamicResizeToolbar

"""
import logging

from AnyQt.QtWidgets import QAction

from AnyQt.QtCore import Qt

from .. import test
from .. import toolbar


class ToolBoxTest(test.QAppTestCase):

    def test_dynamic_toolbar(self):
        logging.basicConfig(level=logging.DEBUG)

        w = toolbar.DynamicResizeToolBar(None)
        w.setStyleSheet("QToolButton { border: 1px solid red; }")

        w.addAction(QAction("1", w))
        w.addAction(QAction("2", w))
        w.addAction(QAction("A long name", w))
        actions = list(w.actions())

        self.assertSequenceEqual([str(action.text()) for action in actions],
                                 ["1", "2", "A long name"])

        w.resize(100, 30)
        w.show()

        w.raise_()

        w.removeAction(actions[1])
        w.insertAction(actions[2], actions[1])

        self.assertSequenceEqual(actions, list(w.actions()),
                                 msg="insertAction does not preserve "
                                     "action order")

        self.singleShot(10, lambda: w.setOrientation(Qt.Vertical))
        self.singleShot(50, lambda: w.removeAction(actions[1]))
        self.qWait()
